from .dual_annealing import DualAnnealing
from .optimizer import Optimizer, Sample, OptimizationFn, OptimizationParams
from .uniform_random import UniformRandom
from .basinhopping import Basinhopping



from __future__ import annotations

import sys
import time
from collections import deque
from typing import Any, Generator, Generic, List, Tuple, TypeVar, Union, Deque

if sys.version_info >= (3, 9):
    from collections.abc import Callable
else:
    from typing import Callable

from numpy.random import default_rng
from typing_extensions import overload, Literal

from .models import (
    Falsification,
    Model,
    ModelResult,
    SimulationParams,
    SimulationResult,
    StaticParameters,
    SignalInterpolators,
)
from .optimizers import Optimizer, Sample, OptimizationFn, OptimizationParams
from .options import Options
from .signals import SignalInterpolator
from .specification import Specification, SpecificationFactory
from .results import Iteration, Result, TimedIteration, Run, TimedRun, TimedResult

_RT = TypeVar("_RT")
_ET = TypeVar("_ET")
_SpecificationOrFactory = Union[Specification, SpecificationFactory]


def _static_params(sample: Sample, options: Options) -> StaticParameters:
    return sample[0 : len(options.static_parameters)]  # type: ignore


def _interpolators(sample: Sample, options: Options) -> SignalInterpolators:
    start = len(options.static_parameters)
    interpolators: List[SignalInterpolator] = []

    for signal in options.signals:
        factory = signal.factory
        end = start + signal.control_points
        interpolators.append(factory.create(signal.interval.astuple(), sample[start:end]))
        start = end

    return interpolators


def _time(fn: Callable[[], _RT]) -> Tuple[float, _RT]:
    t_start = time.perf_counter()
    result = fn()
    t_stop = time.perf_counter()

    return t_stop - t_start, result


def _make_spec(sample: Sample, spec: _SpecificationOrFactory) -> Specification:
    if isinstance(spec, Specification):
        return spec
    else:
        return spec(sample)


def _result_cost(result: ModelResult[Any], spec: Specification) -> float:
    if isinstance(result, (SimulationResult, Falsification)):
        return result.eval_using(spec)
    else:
        raise TypeError(f"Unknown result type {type(result)}")


class CostFn(OptimizationFn, Generic[_ET]):
    def __init__(self, model: Model[_ET], specification: _SpecificationOrFactory, options: Options):
        self.model = model
        self.spec = specification
        self.options = options
        self._iterations: Deque[Iteration[_ET]] = deque()

    def __call__(self, sample: Sample) -> float:
        static_params = _static_params(sample, self.options)
        interpolators = _interpolators(sample, self.options)
        spec = _make_spec(sample, self.spec)

        model_params = SimulationParams(static_params, interpolators, self.options.interval)
        model_result = self.model.simulate(model_params)
        cost = _result_cost(model_result, spec)

        self._iterations.append(Iteration(cost, sample, model_result.data))

        return cost

    @property
    def iterations(self) -> List[Iteration[_ET]]:
        return list(self._iterations)


class TimedCostFn(OptimizationFn, Generic[_ET]):
    def __init__(self, model: Model[_ET], specification: _SpecificationOrFactory, options: Options):
        self.model = model
        self.spec = specification
        self.options = options
        self._iterations: Deque[TimedIteration[_ET]] = deque()

    def __call__(self, sample: Sample) -> float:
        static_params = _static_params(sample, self.options)
        interpolators = _interpolators(sample, self.options)
        spec = _make_spec(sample, self.spec)

        model_params = SimulationParams(static_params, interpolators, self.options.interval)
        model_duration, model_result = _time(lambda: self.model.simulate(model_params))
        cost_duration, cost = _time(lambda: _result_cost(model_result, spec))

        self._iterations.append(
            TimedIteration(cost, sample, model_result.data, model_duration, cost_duration)
        )

        return cost

    @property
    def iterations(self) -> List[TimedIteration[_ET]]:
        return list(self._iterations)


ScenarioResult = Union[Result[_RT, _ET], TimedResult[_RT, _ET]]


class Scenario(Generic[_ET]):
    def __init__(self, model: Model[_ET], specification: _SpecificationOrFactory, options: Options):
        self.model = model
        self.spec = specification
        self.options = options

    @property
    def _params(self) -> Generator[OptimizationParams, None, None]:
        rng = default_rng(self.options.seed)

        for _ in range(self.options.runs):
            run_seed = rng.integers(low=0, high=sys.maxsize)

            yield OptimizationParams(
                self.options.bounds,
                self.options.iterations,
                self.options.behavior,
                run_seed,
            )

    def _run(self, optimizer: Optimizer[_RT]) -> Result[_RT, _ET]:
        def run(params: OptimizationParams) -> Run[_RT, _ET]:
            func = CostFn(self.model, self.spec, self.options)
            duration, result = _time(lambda: optimizer.optimize(func, params))

            return Run(result, func.iterations, duration)

        return Result([run(params) for params in self._params], self.options)

    def _run_timed(self, optimizer: Optimizer[_RT]) -> TimedResult[_RT, _ET]:
        def timedrun(params: OptimizationParams) -> TimedRun[_RT, _ET]:
            func = TimedCostFn(self.model, self.spec, self.options)
            duration, result = _time(lambda: optimizer.optimize(func, params))

            return TimedRun(result, func.iterations, duration)

        return TimedResult([timedrun(params) for params in self._params], self.options)

    @overload
    def run(self, optimizer: Optimizer[_RT], *, timed: Literal[True]) -> TimedResult[_RT, _ET]:
        ...

    @overload
    def run(self, optimizer: Optimizer[_RT], *, timed: Literal[False] = ...) -> Result[_RT, _ET]:
        ...

    def run(self, optimizer: Optimizer[_RT], *, timed: bool = False) -> ScenarioResult[_RT, _ET]:
        if timed:
            return self._run_timed(optimizer)
        else:
            return self._run(optimizer)



from __future__ import annotations

import statistics as stats
import sys
from typing import Dict, Union, NamedTuple

if sys.version_info >= (3, 9):
    from collections.abc import Iterable, Callable
else:
    from typing import Iterable, Callable

import numpy as np
from numpy.typing import NDArray
from typing_extensions import Protocol, Literal, runtime_checkable

from .parser import parse
from .optimizers import Sample

_NumericArray = Union[NDArray[np.float_], NDArray[np.int_]]


@runtime_checkable
class Specification(Protocol):
    def evaluate(self, trajectories: _NumericArray, timestamps: _NumericArray) -> float:
        ...


SpecificationFactory = Callable[[Sample], Specification]

PredicateName = str
PredicateDTypes = Literal["float", "float32", "float64"]


class PredicateProps(NamedTuple):
    column: int
    dtype: PredicateDTypes = "float"


class TLTK(Specification):
    """TLTk specification backend.

    Attributes:
        phi: The specification requirement
        predicates: A set of Predicate(s) used in the requirement
    """

    def __init__(self, phi: str, predicate_props: Dict[PredicateName, PredicateProps]):
        predicate_names = predicate_props.keys()
        parsed = parse(phi, list(predicate_names))

        if parsed is None:
            raise RuntimeError("Could not parse requirement")

        self.tltk_obj = parsed
        self.props = predicate_props

    def evaluate(self, trajectories: _NumericArray, timestamps: _NumericArray) -> float:
        prop_map = self.props.items()
        traces = {name: trajectories[props.column].astype(props.dtype) for name, props in prop_map}

        self.tltk_obj.reset()
        self.tltk_obj.eval_interval(traces, np.array(timestamps, dtype=np.float32))

        return self.tltk_obj.robustness


def _step_widths(times: _NumericArray) -> Iterable[float]:
    """Compute the distance between adjacent elements."""

    for i in range(len(times) - 2):
        yield abs(times[i] - times[i + 1])


class RTAMTDiscrete(Specification):
    """RTAMT discrete-time specification backend.

    Attributes:
        phi: The specification requirement
        predicates: A set of Predicate(s) used in the requirement
    """

    def __init__(self, phi: str, predicate_props: Dict[PredicateName, PredicateProps]):
        try:
            from rtamt import STLDiscreteTimeSpecification, Semantics
        except ModuleNotFoundError:
            raise RuntimeError("RTAMT must be installed to use associated backends")

        if "time" in predicate_props:
            raise ValueError("'time' cannot be used as a predicate name for RTAMT")

        self.rtamt_obj = STLDiscreteTimeSpecification(Semantics.STANDARD)

        self.rtamt_obj.spec = phi
        self.props = {name: props.column for name, props in predicate_props.items()}

        for name, options in predicate_props.items():
            self.rtamt_obj.declare_var(name, options.dtype)

    def evaluate(self, trajectories: _NumericArray, timestamps: _NumericArray) -> float:
        from rtamt import LTLPastifyException

        self.rtamt_obj.reset()

        period = stats.mean(_step_widths(timestamps))
        self.rtamt_obj.set_sampling_period(round(period, 2), "s", 0.1)

        # parse AFTER declaring variables and setting sampling period
        self.rtamt_obj.parse()

        try:
            self.rtamt_obj.pastify()
        except LTLPastifyException:
            pass

        traces = {"time": timestamps.tolist()}
        for name, column in self.props.items():
            traces[name] = trajectories[column].tolist()

        # traces: Dict['time': timestamps, 'variable'(s): trajectories]
        robustness = self.rtamt_obj.evaluate(traces)

        return robustness[-1][1]


class RTAMTDense(Specification):
    """RTAMT dense-time specification backend.

    Attributes:
        phi: The specification requirement
        predicates: A set of Predicate(s) used in the requirement
    """

    def __init__(self, phi: str, predicate_props: Dict[PredicateName, PredicateProps]):
        try:
            from rtamt import STLDenseTimeSpecification, Semantics
        except ModuleNotFoundError:
            raise RuntimeError("RTAMT library must be installed to use RTAMT dense time backend")

        self.rtamt_obj = STLDenseTimeSpecification(Semantics.STANDARD)

        self.props = {name: predicate.column for name, predicate in predicate_props.items()}
        self.rtamt_obj.spec = phi

        for name, options in predicate_props.items():
            self.rtamt_obj.declare_var(name, options.dtype)

    def evaluate(self, trajectories: _NumericArray, timestamps: _NumericArray) -> float:
        from rtamt import LTLPastifyException

        self.rtamt_obj.reset()

        # parse AFTER declaring variables
        self.rtamt_obj.parse()

        try:
            self.rtamt_obj.pastify()
        except LTLPastifyException:
            pass

        column_map = self.props.items()
        traces = [
            (name, np.array([timestamps, trajectories[col]]).T.tolist()) for name, col in column_map
        ]

        # traces: List[Tuple[name, List[Tuple[timestamp, trajectory]]]
        robustness = self.rtamt_obj.evaluate(*traces)

        return robustness[-1][1]



from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from typing import Tuple, Union

if sys.version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence


import numpy as np
from numpy.typing import NDArray
from scipy.interpolate import PchipInterpolator, interp1d
from typing_extensions import Protocol, runtime_checkable, overload

_RealVector = Union[NDArray[np.int_], NDArray[np.float_]]


class SignalInterpolator(Protocol):
    """Interface for callable that provides interpolated signal values at given time t.

    A signal interpolator should provide an interpolated value for the signal at any time t within
    the simulation time interval. All signal interpolators should return their value synchonously.
    """

    @overload
    def interpolate(self, __value: float) -> float:
        ...

    @overload
    def interpolate(
        self, __value: Union[Sequence[float], _RealVector]
    ) -> Union[Sequence[float], _RealVector]:
        ...


@runtime_checkable
class InterpolatorFactory(Protocol):
    """Interface for factory function to return signal interpolators."""

    def create(self, interval: Tuple[float, float], y_values: _RealVector) -> SignalInterpolator:
        ...


class _ScipyFactory(ABC, InterpolatorFactory):
    """Common base class for all interpolator generators wrapping scipy interpolators."""

    @abstractmethod
    def _create(self, x_values: _RealVector, y_values: _RealVector) -> SignalInterpolator:
        raise NotImplementedError()

    def _x_values(self, interval: Tuple[float, float], size: float) -> _RealVector:
        return np.linspace(interval[0], interval[1], size, endpoint=True)  # type: ignore

    def create(self, interval: Tuple[float, float], y_values: _RealVector) -> SignalInterpolator:
        x_values = self._x_values(interval, y_values.size)
        return self._create(x_values, y_values)


_Interpolation = Union[float, _RealVector]


class PchipWrapper(SignalInterpolator):
    def __init__(self, x_values: _RealVector, y_values: _RealVector):
        self.interpolator = PchipInterpolator(x_values, y_values)

    @overload
    def interpolate(self, x: float) -> float:
        ...

    @overload
    def interpolate(self, x: Union[Sequence[float], _RealVector]) -> _RealVector:
        ...

    def interpolate(self, x: Union[float, _RealVector, Sequence[float]]) -> _Interpolation:
        return self.interpolator(x)


class PchipFactory(_ScipyFactory):
    """Factory to return pchip interpolators."""

    def _create(self, x_values: _RealVector, y_values: _RealVector) -> PchipWrapper:
        return PchipWrapper(x_values, y_values)


class LinearWrapper(SignalInterpolator):
    def __init__(self, x_values: _RealVector, y_values: _RealVector):
        self.interpolater = interp1d(x_values, y_values)

    @overload
    def interpolate(self, x: float) -> float:
        ...

    @overload
    def interpolate(self, x: Union[Sequence[float], _RealVector]) -> _RealVector:
        ...

    def interpolate(self, x: Union[float, _RealVector, Sequence[float]]) -> _Interpolation:
        return self.interpolater(x)


class LinearFactory(_ScipyFactory):
    """Factory to create piecewise linear interpolators."""

    def _create(self, x_values: _RealVector, y_values: _RealVector) -> LinearWrapper:
        return LinearWrapper(x_values, y_values)


class ConstantWrapper(SignalInterpolator):
    def __init__(self, x_values: _RealVector, y_values: _RealVector):
        self.interpolater = interp1d(x_values, y_values, kind="zero", fill_value="extrapolate")

    @overload
    def interpolate(self, x: float) -> float:
        ...

    @overload
    def interpolate(self, x: Union[Sequence[float], _RealVector]) -> _RealVector:
        ...

    def interpolate(self, x: Union[float, _RealVector, Sequence[float]]) -> _Interpolation:
        return self.interpolater(x)


class ConstantFactory(_ScipyFactory):
    """Factory to create piecewise constant interpolators."""

    def _create(self, x_values: _RealVector, y_values: _RealVector) -> ConstantWrapper:
        return ConstantWrapper(x_values, y_values)



from __future__ import annotations

import os
import sys
from itertools import takewhile
from typing import Union

if sys.version_info >= (3, 9):
    from collections.abc import Sequence
else:
    from typing import Sequence

import numpy as np
from pathos.multiprocessing import ProcessPool
from numpy.random import default_rng, Generator
from numpy.typing import NDArray
from typing_extensions import Literal

from .optimizer import Optimizer, OptimizationFn, OptimizationParams
from ..options import Interval, Behavior


def _sample(bounds: Sequence[Interval], rng: Generator) -> NDArray[np.float_]:
    return np.array([rng.uniform(bound.lower, bound.upper) for bound in bounds])


class UniformRandom(Optimizer[None]):
    def __init__(self, parallelization: Union[Literal["cores"], int, None] = None):
        if parallelization is not None:
            self.processes = parallelization if isinstance(parallelization, int) else os.cpu_count()
        else:
            self.processes = None

    def optimize(self, func: OptimizationFn, params: OptimizationParams) -> None:
        rng = default_rng(params.seed)
        samples = [_sample(params.bounds, rng) for _ in range(params.iterations)]

        if params.behavior is Behavior.MINIMIZATION:
            if self.processes is not None:
                pool = ProcessPool(nodes=self.processes)
                pool.map(func, samples)
            else:
                for _ in map(func, samples):
                    pass
        else:
            for _ in takewhile(lambda s: func(s) > 0, samples):
                pass



from .staliro import staliro, staliro_timed
from .options import Options, SignalOptions, Behavior
from .scenarios import Scenario
from .specification import TLTK, RTAMTDense, RTAMTDiscrete, PredicateProps
from .models import blackbox, ode
from .optimizers import UniformRandom, DualAnnealing



from __future__ import annotations

from importlib.util import find_spec
from typing import Dict, Optional, TYPE_CHECKING, Union, Sequence

from antlr4.CommonTokenStream import CommonTokenStream
from antlr4.InputStream import InputStream

if TYPE_CHECKING:
    from tltk_mtl import Predicate

    from .stlSpecification import StlSpecification

    PredicateDict = Dict[str, Predicate]
    Predicates = Union[Sequence[str], PredicateDict]


def parse(formula: str, predicates: Predicates, mode: str = "cpu") -> Optional[StlSpecification]:
    """TLTk parser parses a specification requirement into an equivalent TLTk structure

    Attributes:
        formula: The formal specification requirement
        predicates: The set of Predicate(s) used in the requirement
        mode: The TLTk computation mode
    """

    if find_spec("tltk_mtl") is None:
        raise RuntimeError("TLTK must be installed to use parser functionality")

    from .stlLexer import stlLexer as Lexer
    from .stlParser import stlParser as Parser
    from .stlParserVisitor import stlParserVisitor as Visitor

    input_stream = InputStream(formula)

    lexer = Lexer(input_stream)
    stream = CommonTokenStream(lexer)

    parser = Parser(stream)
    tree = parser.stlSpecification()
    visitor = Visitor(lexer, predicates, mode)

    return visitor.visit(tree)  # type: ignore



from __future__ import annotations

from .optimizer import Optimizer, OptimizationFn, OptimizationParams


class Basinhopping(Optimizer[None]):
    def optimize(self, func: OptimizationFn, params: OptimizationParams) -> None:
        raise NotImplementedError()



from __future__ import annotations

import logging
from typing import TypeVar, Union

from .models import Model
from .options import Options
from .optimizers import Optimizer
from .results import Result, TimedResult
from .scenarios import Scenario
from .specification import Specification, SpecificationFactory

logger = logging.getLogger("staliro")
logger.addHandler(logging.NullHandler())

_RT = TypeVar("_RT")
_ET = TypeVar("_ET")


def staliro(
    model: Model[_ET],
    specification: Union[Specification, SpecificationFactory],
    optimizer: Optimizer[_RT],
    options: Options,
) -> Result[_RT, _ET]:
    """Search for falsifying inputs to the provided system.

    Using the optimizer, search the input space defined in the options for cases which falsify the
    provided specification represented by the formula phi and the predicates.

    Args:
        model: The model of the system being tested
        specification: The requirement for the system
        options: General options to manipulate the behavior of the search
        optimizer_cls: The class of the optimizer to use to search the sample space

    Returns:
        results: A list of result objects corresponding to each run from the optimizer
    """

    return Scenario(model, specification, options).run(optimizer)


def staliro_timed(
    model: Model[_ET],
    specification: Union[Specification, SpecificationFactory],
    optimizer: Optimizer[_RT],
    options: Options,
) -> TimedResult[_RT, _ET]:
    """Search for falsifying inputs to the provided system and time system components.

    Using the optimizer, search the input space defined in the options for cases which falsify the
    provided specification represented by the formula phi and the predicates. The time for each
    simulation and each cost computation is stored during execution.

    Args:
        model: The model of the system being tested
        specification: The requirement for the system
        options: General options to manipulate the behavior of the search
        optimizer_cls: The class of the optimizer to use to search the sample space

    Returns:
        results: A list of result objects corresponding to each run from the optimizer
    """

    return Scenario(model, specification, options).run(optimizer, timed=True)



from __future__ import annotations

import math
import sys
from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar, Union, Optional, Type, TYPE_CHECKING

if sys.version_info >= (3, 9):
    from collections.abc import Sequence, Callable
else:
    from typing import Sequence, Callable

import numpy as np
from attr import Attribute, attrs, attrib
from numpy.typing import NDArray
from scipy import integrate
from typing_extensions import overload

from .options import Interval
from .signals import SignalInterpolator
from .specification import Specification

if TYPE_CHECKING:
    _Validator = Callable[[Any, Attribute[Any], Any], Any]


def _ndarray_validator(dims: Sequence[int], dtypes: Sequence[Type[np.generic]]) -> _Validator:
    def validator(inst: Any, attr: Attribute[Any], value: Any) -> None:
        if not isinstance(value, np.ndarray):
            raise TypeError("timestamps may only be of type np.ndarray")

        if not any(value.ndim == dim for dim in dims):
            raise ValueError(f"{attr.name} must be have dimension: {dims}")

        if not any(np.issubdtype(value.dtype, dtype) for dtype in dtypes):
            raise TypeError(f"{attr.name} must have dtype: {dtypes}")

    return validator


_RealVector = Union[NDArray[np.float_], NDArray[np.int_]]
_T = TypeVar("_T")
_numeric_types = (np.integer, np.floating)
_timestamp_validator = _ndarray_validator((1,), _numeric_types)
_trajectories_validator = _ndarray_validator((1, 2), _numeric_types)


class Evaluable(ABC):
    @abstractmethod
    def eval_using(self, spec: Specification) -> float:
        raise NotImplementedError()


@attrs(auto_attribs=True, frozen=True, init=False)
class SimulationResult(Generic[_T], Evaluable):
    _trajectories: _RealVector = attrib(validator=_trajectories_validator, converter=np.array)
    timestamps: _RealVector = attrib(validator=_timestamp_validator, converter=np.array)
    data: _T

    @overload
    def __init__(self: SimulationResult[None], trajectories: _RealVector, timestamps: _RealVector):
        ...

    @overload
    def __init__(self, trajectories: _RealVector, timestamps: _RealVector, data: _T):
        ...

    def __init__(self, trajectories: _RealVector, timestamps: _RealVector, data: _T = None):
        self.__attrs_init__(trajectories, timestamps, data)  # type: ignore

    def eval_using(self, spec: Specification) -> float:
        return spec.evaluate(self.trajectories, self.timestamps)

    @property
    def trajectories(self) -> _RealVector:
        trajectories = np.atleast_2d(self._trajectories)

        if trajectories.shape[0] == self.timestamps.shape[0]:
            return trajectories.T
        elif trajectories.shape[1] == self.timestamps.shape[0]:
            return trajectories
        else:
            raise ValueError("No dimension of the trajectories matrix matches the timestamps len")


@attrs(auto_attribs=True, frozen=True, init=False)
class Falsification(Generic[_T], Evaluable):
    data: _T

    @overload
    def __init__(self: Falsification[None]):
        ...

    @overload
    def __init__(self, data: _T):
        ...

    def __init__(self, data: _T = None):
        self.__attrs_init__(data)  # type: ignore

    def eval_using(self, spec: Specification) -> float:
        return -math.inf


StaticParameters = _RealVector
SignalInterpolators = Sequence[SignalInterpolator]
ModelResult = Union[SimulationResult[_T], Falsification[_T]]


@attrs(auto_attribs=True, frozen=True)
class SimulationParams:
    static_parameters: StaticParameters
    interpolators: SignalInterpolators
    interval: Interval


class Model(ABC, Generic[_T]):
    @abstractmethod
    def simulate(self, params: SimulationParams) -> ModelResult[_T]:
        raise NotImplementedError()


SignalTimes = NDArray[np.float_]
SignalValues = NDArray[np.float_]
Timestamps = Union[_RealVector, Sequence[float], Sequence[int]]
Trajectories = Union[_RealVector, Sequence[Sequence[float]], Sequence[Sequence[int]]]
BlackboxFunc = Callable[[StaticParameters, SignalTimes, SignalValues], ModelResult[_T]]


class Blackbox(Model[_T]):
    def __init__(self, func: BlackboxFunc[_T], sampling_interval: float = 0.1):
        self.func = func
        self.sampling_interval = sampling_interval

    def simulate(self, params: SimulationParams) -> ModelResult[_T]:
        interval = params.interval
        duration = interval.upper - interval.lower
        point_count = math.floor(duration / self.sampling_interval)
        signal_times = np.linspace(start=interval.lower, stop=interval.upper, num=point_count)
        signal_traces = [
            interpolator.interpolate(signal_times) for interpolator in params.interpolators
        ]

        return self.func(params.static_parameters, signal_times, np.array(signal_traces))


Time = float
State = _RealVector
IntegrationFn = Callable[[float, _RealVector], _RealVector]
ODEResult = Union[_RealVector, Sequence[float], Sequence[int]]
ODEFunc = Callable[[Time, State, SignalValues], ODEResult]


def _make_integration_fn(signals: SignalInterpolators, func: ODEFunc) -> IntegrationFn:
    def integration_fn(time: float, state: State) -> State:
        signal_values = [signal.interpolate(time) for signal in signals]
        result = func(time, state, np.array(signal_values))
        return np.array(result)

    return integration_fn


class ODE(Model[None]):
    def __init__(self, func: ODEFunc):
        self.func = func

    def simulate(self, params: SimulationParams) -> SimulationResult[None]:
        integration_fn = _make_integration_fn(params.interpolators, self.func)
        integration = integrate.solve_ivp(
            integration_fn, params.interval.astuple(), params.static_parameters
        )

        return SimulationResult(integration.y, integration.t, None)


_BlackboxDecorator = Callable[[BlackboxFunc[_T]], Blackbox[_T]]


@overload
def blackbox(*, sampling_interval: float = ...) -> Callable[[BlackboxFunc[_T]], Blackbox[_T]]:
    ...


@overload
def blackbox(_func: BlackboxFunc[_T]) -> Blackbox[_T]:
    ...


def blackbox(
    _func: Optional[BlackboxFunc[_T]] = None, *, sampling_interval: float = 0.1
) -> Union[Blackbox[_T], _BlackboxDecorator[_T]]:
    """Decorate a function as a blackbox model.

    This decorator can be used with or without arguments.

    Args:
        func: The function to wrap as a blackbox
        *: Variable length argument list
        sampling_interval: The time interval to use when sampling the signal interpolators

    Returns:
        A blackbox model
    """

    def decorator(func: BlackboxFunc[_T]) -> Blackbox[_T]:
        return Blackbox(func, sampling_interval)

    if _func is not None:
        return decorator(_func)
    else:
        return decorator


_ODEDecorator = Callable[[ODEFunc], ODE]


@overload
def ode() -> _ODEDecorator:
    ...


@overload
def ode(_func: ODEFunc) -> ODE:
    ...


def ode(_func: Optional[ODEFunc] = None) -> Union[_ODEDecorator, ODE]:
    def decorator(func: ODEFunc) -> ODE:
        return ODE(func)

    if _func is not None:
        return decorator(_func)
    else:
        return decorator



from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TypeVar, Generic, Sequence, Union

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

from numpy import float_, int_
from numpy.typing import NDArray

from ..options import Interval, Behavior


@dataclass(frozen=True)
class OptimizationParams:
    bounds: Sequence[Interval]
    iterations: int
    behavior: Behavior
    seed: int


Sample = Union[NDArray[float_], NDArray[int_]]


class OptimizationFn(Protocol):
    def __call__(self, __sample: Sample) -> float:
        ...


_T = TypeVar("_T")


class Optimizer(ABC, Generic[_T]):
    @abstractmethod
    def optimize(self, func: OptimizationFn, params: OptimizationParams) -> _T:
        raise NotImplementedError()



from typing import Union

import tltk_mtl as mtl

StlSpecification = Union[
    mtl.And,
    mtl.Finally,
    mtl.Global,
    mtl.Implication,
    mtl.Next,
    mtl.Not,
    mtl.Or,
    mtl.Predicate,
    mtl.Until,
]



from __future__ import annotations

from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray
from scipy import optimize
from typing_extensions import Literal

from .optimizer import Optimizer, OptimizationFn, OptimizationParams, Sample
from ..options import Behavior


@dataclass
class DualAnnealingResult:
    jacobian_value: NDArray[np.float_]
    jacobian_evals: int
    hessian_value: NDArray[np.float_]
    hessian_evals: int


class DualAnnealing(Optimizer[DualAnnealingResult]):
    def optimize(self, func: OptimizationFn, options: OptimizationParams) -> DualAnnealingResult:
        def wrapper(sample: Sample) -> float:
            return func(sample)

        def listener(sample: Sample, robustness: float, ctx: Literal[-1, 0, 1]) -> bool:
            if robustness < 0 and options.behavior is Behavior.FALSIFICATION:
                return True

            return False

        bounds = [bound.astuple() for bound in options.bounds]
        result: optimize.OptimizeResult = optimize.dual_annealing(
            wrapper,
            bounds,
            seed=options.seed,
            maxiter=options.iterations,
            no_local_search=True,  # Disable local search, use only traditional generalized SA
            callback=listener,
        )

        return DualAnnealingResult(result.jac, result.njev, result.hess, result.nhev)



from __future__ import annotations

import statistics as stats
import sys
from dataclasses import dataclass
from typing import Generic, TypeVar

if sys.version_info >= (3, 9):
    from collections.abc import Iterable, Sequence
else:
    from typing import Iterable, Sequence

from .optimizers import Sample
from .options import Options

_ET = TypeVar("_ET")


@dataclass(frozen=True)
class Iteration(Generic[_ET]):
    cost: float
    sample: Sample
    extra: _ET


@dataclass(frozen=True)
class TimedIteration(Iteration[_ET]):
    model_duration: float
    cost_duration: float


_RT = TypeVar("_RT")


@dataclass(frozen=True)
class Run(Generic[_RT, _ET]):
    result: _RT
    history: Sequence[Iteration[_ET]]
    duration: float

    @property
    def best_iter(self) -> Iteration[_ET]:
        return min(self.history, key=lambda i: i.cost)


@dataclass(frozen=True)
class TimeStats:
    _durations: Iterable[float]

    @property
    def average_time(self) -> float:
        return stats.mean(self._durations)

    @property
    def total_time(self) -> float:
        return sum(self._durations)

    @property
    def longest_duration(self) -> float:
        return max(self._durations)


@dataclass(frozen=True)
class TimedRun(Run[_RT, _ET]):
    history: Sequence[TimedIteration[_ET]]

    @property
    def best_iter(self) -> TimedIteration[_ET]:
        return min(self.history, key=lambda i: i.cost)

    @property
    def fastest_iter(self) -> TimedIteration[_ET]:
        return min(self.history, key=lambda i: i.model_duration + i.cost_duration)

    @property
    def model(self) -> TimeStats:
        return TimeStats(iteration.model_duration for iteration in self.history)

    @property
    def cost_fn(self) -> TimeStats:
        return TimeStats(iteration.cost_duration for iteration in self.history)


@dataclass(frozen=True)
class Result(Generic[_RT, _ET]):
    runs: Sequence[Run[_RT, _ET]]
    options: Options

    @property
    def best_run(self) -> Run[_RT, _ET]:
        return min(self.runs, key=lambda r: r.best_iter.cost)


@dataclass(frozen=True)
class TimedResult(Result[_RT, _ET]):
    runs: Sequence[TimedRun[_RT, _ET]]

    @property
    def best_run(self) -> TimedRun[_RT, _ET]:
        return min(self.runs, key=lambda r: r.best_iter.cost)



from __future__ import annotations

import enum
import random
import sys
from typing import Any, List, Optional, Tuple, Union

if sys.version_info >= (3, 9):
    from collections.abc import Iterable, Sequence
else:
    from typing import Iterable, Sequence

from attr import attrs, attrib, Attribute
from attr.validators import optional
from typing_extensions import Literal

from .signals import InterpolatorFactory, PchipFactory


def _bounds_converter(value: Any) -> Any:
    if isinstance(value, Interval):
        return value.bounds

    return value


@attrs(auto_attribs=True)
class Interval:
    bounds: Sequence[float] = attrib(converter=_bounds_converter)

    @bounds.validator
    def _validate_values(self, attr: Attribute[Sequence[float]], value: Sequence[float]) -> None:
        if len(value) != 2:
            raise ValueError("only 2 values may be passed to interval")

        if value[0] > value[1]:
            raise ValueError("first value must be strictly less than second value")

    @property
    def lower(self) -> float:
        return self.bounds[0]

    @property
    def upper(self) -> float:
        return self.bounds[1]

    def astuple(self) -> Tuple[float, float]:
        return (self.lower, self.upper)


SignalTimes = Sequence[float]


@attrs
class SignalOptions:
    """Options for signal generation.

    Attributes:
        bound: The interval the signal should be generated for
        control_points: The number of points the optimizer should generate for the signal
        factory: Factory to produce interpolators for the signal
        step: The step size when evaluating the signal
        signal_times: Time values for the generated control points
        time_varying: Flag that indicates that the signal times should be considered a search
                      variable (EXPERIMENTAL)
    """

    interval: Interval = attrib(converter=Interval)
    factory: InterpolatorFactory = attrib(factory=PchipFactory)
    control_points: int = attrib(default=10, converter=int)
    step: float = attrib(default=0.1, converter=float)
    signal_times: Optional[SignalTimes] = attrib(default=None)
    time_varying: bool = False  # Boolean flag for turning control point times into search variables

    @control_points.validator
    def _validate_control_points(self, attr: Attribute[int], value: int) -> None:
        if value < 0:
            raise ValueError("control points must be greater than zero")

    @step.validator
    def _validate_step(self, attr: Attribute[float], value: float) -> None:
        if value < 0:
            raise ValueError("step must be greater than zero")

    @factory.validator
    def _validate_factory(
        self, attr: Attribute[InterpolatorFactory], value: InterpolatorFactory
    ) -> None:
        if not isinstance(value, InterpolatorFactory):
            raise ValueError("factory must implement InterpolatorFactory protocol")

    @signal_times.validator
    def _validate_signal_times(
        self, attr: Attribute[Optional[SignalTimes]], value: Optional[SignalTimes]
    ) -> None:
        if value is not None and len(value) != self.control_points:
            raise ValueError("must specify as many signal times as control points")

    @property
    def bounds(self) -> List[Interval]:
        return [self.interval] * self.control_points


class Behavior(enum.IntEnum):
    """Behavior when falsifying case for system is encountered.

    Attributes:
        FALSIFICATION: Stop searching when the first falsifying case is encountered
        MINIMIZATION: Continue searching after encountering a falsifying case until iteration
                      budget is exhausted
    """

    FALSIFICATION = enum.auto()
    MINIMIZATION = enum.auto()


def _static_parameter_converter(obj: Any) -> List[Interval]:
    return [Interval(elem) for elem in obj]


def _seed_factory() -> int:
    return random.randint(0, sys.maxsize)


_ParallelizationT = Union[int, Literal["all", "cores"], None]


def _parallelization_validator(
    _: Any, attr: Attribute[_ParallelizationT], value: _ParallelizationT
) -> None:
    if isinstance(value, str) and value != "cores" and value != "all":
        raise ValueError()
    elif not isinstance(value, int):
        raise TypeError()


@attrs
class Options:
    """General options for controlling falsification behavior.

    Attributes:
        iterations: The number of search iterations to perform in a run
        runs: The number times to run the optimizer
        interval: The time interval of the system simulation
        behavior: The behavior of the system when a falsifying case is found
        static_parameters: Parameters that will be provided to the system at the beginning and are
                           time invariant (initial conditions)
        signals: System inputs that will vary over time
        verbose: Print additional data during execution
        bounds: The combined bounds from both the static_parameters and signals
    """

    static_parameters: List[Interval] = attrib(factory=list, converter=_static_parameter_converter)
    signals: Iterable[SignalOptions] = attrib(factory=list)
    seed: int = attrib(factory=_seed_factory)
    iterations: int = attrib(default=400, converter=int)
    runs: int = attrib(default=1, converter=int)
    interval: Interval = attrib(default=Interval([0, 1]), converter=Interval)
    sampling_interval: float = attrib(default=0.1, converter=float)
    behavior: Behavior = attrib(default=Behavior.FALSIFICATION)
    parallelization: _ParallelizationT = attrib(
        default=None, validator=optional(_parallelization_validator)
    )
    verbose: bool = False

    @runs.validator
    def _validate_runs(self, attr: Attribute[int], value: int) -> None:
        if value < 0:
            raise ValueError("runs must be greater than zero")

    @sampling_interval.validator
    def _validate_sampling_interval(self, attr: Attribute[float], value: float) -> None:
        if value < 0:
            raise ValueError("sampling interval must be greater than zero")

    @signals.validator
    def _validate_signals(
        self, attr: Attribute[List[SignalOptions]], signals: List[SignalOptions]
    ) -> None:
        for signal in signals:
            if not isinstance(signal, SignalOptions):
                raise ValueError("can only provide SignalOptions objects to signals attribute")

    @property
    def bounds(self) -> List[Interval]:
        signal_bounds = [bound for signal in self.signals for bound in signal.bounds]
        static_bounds = list(self.static_parameters)

        return static_bounds + signal_bounds
